class RustEntityInfo:

    def __init__(self, data) -> None:
        self.type: int = data.type
        self.value: bool = data.payload.value
        self.items = [RustEntityInfoItem(item) for item in data.payload.items]
        self.capacity: int = data.payload.capacity
        self.hasProtection: bool = data.payload.hasProtection
        self.protectionExpiry: int = data.payload.protectionExpiry

    def __str__(self) -> str:
        return "RustEntityInfo[type={}, value={}, items={}, capacity={}, hasProtection={}, protectionExpiry={}]".format(
            self.type, self.value, self.items, self.capacity, self.hasProtection, self.protectionExpiry)


class RustEntityInfoItem:

    def __init__(self, data) -> None:
        self.itemId: int = data.itemId
        self.quantity: int = data.quantity
        self.itemIsBlueprint: bool = data.itemIsBlueprint

    def __str__(self) -> str:
        return "RustEntityInfoItem[itemId={}, quantity={}, itemIsBlueprint={}]".format(self.itemId, self.quantity,
                                                                                       self.itemIsBlueprint)



class RustItem:

    def __init__(self, name: str, item_id: int, quantity: int, is_blueprint: bool) -> None:
        self.name: str = name
        self.itemId: int = item_id
        self.quantity: int = quantity
        self.isBlueprint: bool = is_blueprint

    def __str__(self) -> str:
        return "RustItem[name={}, itemId={}, quantity={}, isBlueprint={}]".format(self.name, self.itemId, self.quantity,
                                                                                  self.isBlueprint)



class RustMap:

    def __init__(self, data) -> None:
        self.width: int = data.width
        self.height: int = data.height
        self.jpgImage: bytes = data.jpgImage
        self.margin: int = data.oceanMargin
        self.monuments = [RustMonument(monument.token, monument.x, monument.y) for monument in data.monuments]
        self.background: str = None if data.background == "" else data.background

    def __str__(self) -> str:
        return "RustMap[width={}, height={}, jpgImage={}, margin={}, monuments={}, background={}]".format(self.width,
                                                                                                          self.height,
                                                                                                          len(self.jpgImage),
                                                                                                          self.margin,
                                                                                                          self.monuments,
                                                                                                          self.background)


class RustMonument:

    def __init__(self, token, x, y) -> None:
        self.token: str = token
        self.x: float = x
        self.y: float = y

    def __str__(self) -> str:
        return "RustMonument[token={}, x={}, y={}]"



class RustMarker:

    def __init__(self, data) -> None:
        self.id: int = data.id
        self.type: int = data.type
        self.x: float = data.x
        self.y: float = data.y
        self.steamId: int = data.steamId
        self.rotation: float = data.rotation
        self.radius: float = data.radius
        self.colour1 = RustColour(data.color1)
        self.colour2 = RustColour(data.color2)
        self.alpha: float = data.alpha
        self.name: str = data.name
        self.sellOrders = [RustSellOrder(order) for order in data.sellOrders]

    def __str__(self) -> str:
        return "RustMarker[id={}, type={}, x={}, y={}, steamId={}, rotation={}, radius={}, colour1={}, colour2={}, " \
               "alpha={}, name={}, sellOrders={}]".format(
                    self.id, self.type, self.x, self.y, self.steamId, self.rotation, self.radius, self.colour1,
                    self.colour2, self.alpha, self.name, self.sellOrders)


class RustColour:

    def __init__(self, data) -> None:
        self.x: float = data.x
        self.y: float = data.y
        self.z: float = data.z
        self.w: float = data.w

    def __str__(self) -> str:
        return "RustColour[x={}, y={}, z={}, w={}]".format(self.x, self.y, self.z, self.w)


class RustSellOrder:

    def __init__(self, data) -> None:
        self.itemId: int = data.itemId
        self.quantity: int = data.quantity
        self.currencyId: int = data.currencyId
        self.costPerItem: int = data.costPerItem
        self.itemIsBlueprint: bool = data.itemIsBlueprint
        self.currencyIsBlueprint: bool = data.currencyIsBlueprint
        self.amountInStock: int = data.amountInStock

    def __str__(self) -> str:
        return "RustSellOrder[itemId={}, quantity={}, currencyId={}, costPerItem={}, itemIsBlueprint={}, " \
               "currencyIsBlueprint={}]".format(
                    self.itemId, self.quantity, self.currencyId, self.costPerItem, self.itemIsBlueprint,
                    self.currencyIsBlueprint)



import asyncio
from asyncio.futures import Future
from typing import List
from PIL import Image

from .structures import *
from .remote.rustplus_pb2 import *
from .remote import RustRemote, HeartBeat
from ..commands import CommandOptions
from ..exceptions import *
from ..utils import RegisteredListener


class BaseRustSocket:

    def __init__(self, ip: str = None, port: str = None, steamid: int = None, playertoken: int = None,
                 command_options: CommandOptions = None, raise_ratelimit_exception: bool = True,
                 ratelimit_limit: int = 25, ratelimit_refill: int = 3, heartbeat: HeartBeat = None, use_proxy: bool = False) -> None:

        if ip is None:
            raise ValueError("Ip cannot be None")
        if port is None:
            raise ValueError("Port cannot be None")
        if steamid is None:
            raise ValueError("SteamID cannot be None")
        if playertoken is None:
            raise ValueError("PlayerToken cannot be None")

        self.ip = ip
        self.port = port
        self.steamid = steamid
        self.player_token = playertoken
        self.seq = 1
        self.command_options = command_options
        self.raise_ratelimit_exception = raise_ratelimit_exception
        self.listener_seq = 0

        self.remote = RustRemote(ip=self.ip, port=self.port, command_options=command_options,
                                 ratelimit_limit=ratelimit_limit, ratelimit_refill=ratelimit_refill, use_proxy=use_proxy)

        if heartbeat is None:
            raise ValueError("Heartbeat cannot be None")
        self.heartbeat = heartbeat

    async def _handle_ratelimit(self, amount=1) -> None:
        """
        Handles the ratelimit for a specific request
        """

        while True:

            if self.remote.ratelimiter.can_consume(amount):
                self.remote.ratelimiter.consume(amount)
                self.heartbeat.reset_rhythm()
                return

            if self.raise_ratelimit_exception:
                raise RateLimitError("Out of tokens")

            await asyncio.sleep(self.remote.ratelimiter.get_estimated_delay_time(amount))

    def _generate_protobuf(self) -> AppRequest:
        """
        Generates the empty AppRequest Message
        """

        app_request = AppRequest()
        app_request.seq = self.seq
        app_request.playerId = self.steamid
        app_request.playerToken = self.player_token

        self.seq += 1

        return app_request

    async def connect(self) -> None:
        """
        Opens the connection to the Rust Server
        """
        try:
            if self.remote.ws is None:
                self.remote.connect()
                await self.send_wakeup_request()
                await self.heartbeat.start_beat()
        except ConnectionRefusedError:
            raise ServerNotResponsiveError("Cannot Connect")

    async def close_connection(self) -> None:
        """
        Disconnects from the Rust Server
        """
        self.remote.close()

    async def disconnect(self) -> None:
        """
        Disconnects from the Rust Server
        """
        await self.close_connection()

    async def send_wakeup_request(self) -> None:
        """
        Sends a request to the server to wake up broadcast responses
        """
        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getTime.CopyFrom(AppEmpty())

        self.remote.ignored_responses.append(app_request.seq)

        await self.remote.send_message(app_request)

    def command(self, coro) -> RegisteredListener:
        """
        A Decorator to register a command listener
        """

        if self.command_options is None:
            raise CommandsNotEnabledError("Not enabled")

        if isinstance(coro, RegisteredListener):
            coro = coro.get_coro()

        data = (coro, asyncio.get_event_loop())
        self.remote.command_handler.register_command(coro.__name__, data)
        return RegisteredListener(coro.__name__, data)

    def team_event(self, coro) -> RegisteredListener:
        """
        A Decorator to register an event listener for team changes
        """

        if isinstance(coro, RegisteredListener):
            coro = coro.get_coro()

        data = (coro, asyncio.get_event_loop())
        self.remote.event_handler.register_event("team_changed", data)
        return RegisteredListener("team_changed", data)

    def chat_event(self, coro) -> RegisteredListener:
        """
        A Decorator to register an event listener for chat messages
        """

        if isinstance(coro, RegisteredListener):
            coro = coro.get_coro()

        data = (coro, asyncio.get_event_loop())
        self.remote.event_handler.register_event("chat_message", data)
        return RegisteredListener("chat_message", data)

    def entity_event(self, eid):
        """
        Decorator to register a smart device listener
        """

        def wrap_func(coro) -> RegisteredListener:

            if isinstance(coro, RegisteredListener):
                coro = coro.get_coro()

            def entity_event_callback(future_inner: Future):
                try:
                    entity_info: RustEntityInfo = future_inner.result()
                    self.remote.event_handler.register_event(eid, (coro, loop, entity_info.type))
                except Exception:
                    raise SmartDeviceRegistrationError("Not Found")

            loop = asyncio.get_event_loop()
            future = asyncio.run_coroutine_threadsafe(self.get_entity_info(eid), loop)
            future.add_done_callback(entity_event_callback)

            return RegisteredListener(eid, coro)

        return wrap_func

    def remove_listener(self, listener) -> bool:
        """
        This will remove a listener, command or event. Takes a RegisteredListener instance
        :returns Success of removal. True = Removed. False = Not Removed
        """

        if isinstance(listener, RegisteredListener):
            return self.remote.remove_listener(listener)
        return False

    @staticmethod
    async def hang() -> None:
        """
        This Will permanently put your script into a state of 'hanging' Cannot be Undone. Only do this in scripts using commands
        """

        while True:
            await asyncio.sleep(1)

    async def get_time(self) -> RustTime:
        """
        Gets the current in-game time from the server. Returns a RustTime object
        """
        raise NotImplementedError("Not Implemented")

    async def send_team_message(self, message: str) -> None:
        """
        Sends a message to the in-game team chat
        """
        raise NotImplementedError("Not Implemented")

    async def get_info(self) -> RustInfo:
        """
        Gets information on the Rust Server
        """
        raise NotImplementedError("Not Implemented")

    async def get_team_chat(self) -> List[RustChatMessage]:
        """
        Gets the team chat from the server
        """
        raise NotImplementedError("Not Implemented")

    async def get_team_info(self):
        """
        Gets Information on the members of your team
        """
        raise NotImplementedError("Not Implemented")

    async def get_markers(self) -> List[RustMarker]:
        """
        Gets all the map markers from the server
        """
        raise NotImplementedError("Not Implemented")

    async def get_map(self, add_icons: bool = False, add_events: bool = False, add_vending_machines: bool = False,
                      override_images: dict = None) -> Image:
        """
        Gets an image of the map from the server with the specified additions
        """
        raise NotImplementedError("Not Implemented")

    async def get_raw_map_data(self) -> RustMap:
        """
        Gets the raw map data from the server
        """
        raise NotImplementedError("Not Implemented")

    async def get_entity_info(self, eid: int = None) -> RustEntityInfo:
        """
        Gets entity info from the server
        """
        raise NotImplementedError("Not Implemented")

    async def turn_on_smart_switch(self, eid: int = None) -> None:
        """
        Turns on a given smart switch by entity ID
        """
        raise NotImplementedError("Not Implemented")

    async def turn_off_smart_switch(self, eid: int = None) -> None:
        """
        Turns off a given smart switch by entity ID
        """
        raise NotImplementedError("Not Implemented")

    async def promote_to_team_leader(self, steamid: int = None) -> None:
        """
        Promotes a given user to the team leader by their 64-bit Steam ID
        """
        raise NotImplementedError("Not Implemented")

    async def get_current_events(self) -> List[RustMarker]:
        """
        Returns all the map markers that are for events:
        Can detect:
            - Explosion
            - CH47 (Chinook)
            - Cargo Ship
            - Locked Crate
            - Attack Helicopter
        """
        raise NotImplementedError("Not Implemented")

    async def get_tc_storage_contents(self, eid: int = None, combine_stacks: bool = False) -> RustContents:
        """
        Gets the Information about TC Upkeep and Contents.
        Do not use this for any other storage monitor than a TC
        """
        raise NotImplementedError("Not Implemented")



from typing import List


class CommandTime:

    def __init__(self, formatted_time, raw_time) -> None:
        self.formatted_time = formatted_time
        self.raw_time = raw_time


class Command:

    def __init__(self, sender_name: str, sender_steamid: int, time: CommandTime, command: str, args: List[str]) -> None:
        self.sender_name = sender_name
        self.sender_steamid = sender_steamid
        self.time = time
        self.command = command
        self.args = args



from typing import List

from .rust_chat_message import RustChatMessage
from .rust_team_info import RustTeamInfo


class EntityEvent:

    def __init__(self, app_message, type) -> None:
        self.type = int(type)
        self.entityId: int = app_message.broadcast.entityChanged.entityId
        self.value: bool = app_message.broadcast.entityChanged.payload.value
        self.capacity: int = app_message.broadcast.entityChanged.payload.capacity
        self.hasProtection: bool = app_message.broadcast.entityChanged.payload.hasProtection
        self.protectionExpiry: int = app_message.broadcast.entityChanged.payload.protectionExpiry

        self.items: List[Item] = [Item(item) for item in app_message.broadcast.entityChanged.payload.items]


class Item:

    def __init__(self, app_message) -> None:
        self.itemId: int = app_message.itemId
        self.quantity: int = app_message.quantity
        self.itemIsBlueprint: bool = app_message.itemIsBlueprint


class TeamEvent:

    def __init__(self, app_message) -> None:
        self.playerId: int = app_message.broadcast.teamChanged.playerId
        self.teamInfo = RustTeamInfo(app_message.broadcast.teamChanged.teamInfo)


class ChatEvent:

    def __init__(self, app_message) -> None:
        self.message = RustChatMessage(app_message.broadcast.teamMessage.message)



from typing import List
from PIL import Image
from io import BytesIO
from datetime import datetime
from collections import defaultdict

from .base_rust_api import BaseRustSocket
from .structures import RustInfo, RustMap, RustMarker, RustChatMessage, RustTeamInfo, RustEntityInfo, RustContents, \
    RustItem
from .remote.rustplus_pb2 import *
from .remote import HeartBeat
from ..commands import CommandOptions
from ..exceptions import *
from ..utils import *


class RustSocket(BaseRustSocket):

    def __init__(self, ip: str = None, port: str = None, steamid: int = None, playertoken: int = None,
                 command_options: CommandOptions = None, raise_ratelimit_exception: bool = True,
                 ratelimit_limit: int = 25, ratelimit_refill: int = 3, use_proxy: bool = False) -> None:
        super().__init__(ip=ip, port=port, steamid=steamid, playertoken=playertoken, command_options=command_options,
                         raise_ratelimit_exception=raise_ratelimit_exception, ratelimit_limit=ratelimit_limit,
                         ratelimit_refill=ratelimit_refill, heartbeat=HeartBeat(self), use_proxy=use_proxy)

    async def get_time(self) -> RustTime:

        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getTime.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        response = await self.remote.get_response(app_request.seq, app_request)

        return format_time(response)

    async def send_team_message(self, message: str) -> None:

        await self._handle_ratelimit(2)

        app_send_message = AppSendMessage()
        app_send_message.message = message

        app_request = self._generate_protobuf()
        app_request.sendTeamMessage.CopyFrom(app_send_message)

        self.remote.ignored_responses.append(app_request.seq)

        await self.remote.send_message(app_request)

    async def get_info(self) -> RustInfo:

        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getInfo.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        response = await self.remote.get_response(app_request.seq, app_request)

        return RustInfo(response.response.info)

    async def get_team_chat(self) -> List[RustChatMessage]:

        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getTeamChat.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        messages = (await self.remote.get_response(app_request.seq, app_request)).response.teamChat.messages

        return [RustChatMessage(message) for message in messages]

    async def get_team_info(self):

        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getTeamInfo.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        app_message = await self.remote.get_response(app_request.seq, app_request)

        return RustTeamInfo(app_message.response.teamInfo)

    async def get_markers(self) -> List[RustMarker]:

        await self._handle_ratelimit()

        app_request = self._generate_protobuf()
        app_request.getMapMarkers.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        app_message = await self.remote.get_response(app_request.seq, app_request)

        return [RustMarker(marker) for marker in app_message.response.mapMarkers.markers]

    async def get_raw_map_data(self) -> RustMap:

        await self._handle_ratelimit(5)

        app_request = self._generate_protobuf()
        app_request.getMap.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        app_message = await self.remote.get_response(app_request.seq, app_request)

        return RustMap(app_message.response.map)

    async def get_map(self, add_icons: bool = False, add_events: bool = False, add_vending_machines: bool = False,
                      override_images: dict = None) -> Image:

        if override_images is None:
            override_images = {}

        map_size = int((await self.get_info()).size)

        await self._handle_ratelimit(5 + 1 if [add_icons, add_events, add_vending_machines].count(True) >= 1 else 0)

        app_request = self._generate_protobuf()
        app_request.getMap.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        app_message = await self.remote.get_response(app_request.seq, app_request)

        game_map = app_message.response.map
        monuments = list(game_map.monuments)

        try:
            image = Image.open(BytesIO(game_map.jpgImage))
        except Exception:
            raise ImageError("Invalid bytes for the image")

        image = image.crop((500, 500, game_map.height - 500, game_map.width - 500))

        game_map = image.resize((map_size, map_size), Image.ANTIALIAS)

        if add_icons or add_events or add_vending_machines:
            map_markers = await self.get_markers()

            if add_icons:
                for monument in monuments:
                    if str(monument.token) == "DungeonBase":
                        continue
                    icon = convert_monument(monument.token, override_images)
                    if monument.token in override_images:
                        icon = icon.resize((150, 150))
                    if str(monument.token) == "train_tunnel_display_name":
                        icon = icon.resize((100, 125))
                    game_map.paste(icon, (format_coord(int(monument.x), int(monument.y), map_size)), icon)

            if add_vending_machines:
                with resources.path("rustplus.api.icons", "vending_machine.png") as path:
                    vending_machine = Image.open(path).convert("RGBA")
                    vending_machine = vending_machine.resize((100, 100))

            for marker in map_markers:
                if add_events:
                    if marker.type == 2 or marker.type == 4 or marker.type == 5 or marker.type == 6 or marker.type == 8:
                        icon = convert_marker(str(marker.type), marker.rotation)
                        if marker.type == 6:
                            x = marker.x
                            y = marker.y
                            if y > map_size: y = map_size
                            if y < 0: y = 100
                            if x > map_size: x = map_size - 75
                            if x < 0: x = 50
                            game_map.paste(icon, (int(x), map_size - int(y)), icon)
                        else:
                            game_map.paste(icon, (format_coord(int(marker.x), int(marker.y), map_size)), icon)
                if add_vending_machines and marker.type == 3:
                    game_map.paste(vending_machine, (int(marker.x) - 50, map_size - int(marker.y) - 50),
                                   vending_machine)

        return game_map.resize((2000, 2000), Image.ANTIALIAS)

    async def get_entity_info(self, eid: int = None) -> RustEntityInfo:

        await self._handle_ratelimit()

        if eid is None:
            raise ValueError("EID cannot be None")

        app_request = self._generate_protobuf()
        app_request.entityId = eid
        app_request.getEntityInfo.CopyFrom(AppEmpty())

        await self.remote.send_message(app_request)

        app_message = await self.remote.get_response(app_request.seq, app_request)

        return RustEntityInfo(app_message.response.entityInfo)

    async def _update_smart_device(self, eid: int, value: bool) -> None:

        await self._handle_ratelimit()

        entity_value = AppSetEntityValue()
        entity_value.value = value

        app_request = self._generate_protobuf()

        app_request.entityId = eid
        app_request.setEntityValue.CopyFrom(entity_value)

        self.remote.ignored_responses.append(app_request.seq)

        await self.remote.send_message(app_request)

    async def turn_on_smart_switch(self, eid: int = None) -> None:

        if eid is None:
            raise ValueError("EID cannot be None")

        await self._update_smart_device(eid, True)

    async def turn_off_smart_switch(self, eid: int = None) -> None:

        if eid is None:
            raise ValueError("EID cannot be None")

        await self._update_smart_device(eid, False)

    async def promote_to_team_leader(self, steamid: int = None) -> None:

        if steamid is None:
            raise ValueError("SteamID cannot be None")

        await self._handle_ratelimit()

        leader_packet = AppPromoteToLeader()
        leader_packet.steamId = steamid

        app_request = self._generate_protobuf()
        app_request.promoteToLeader.CopyFrom(leader_packet)

        self.remote.ignored_responses.append(app_request.seq)

        await self.remote.send_message(app_request)

    async def get_current_events(self) -> List[RustMarker]:

        return [marker for marker in (await self.get_markers()) if
                marker.type == 2 or marker.type == 4 or marker.type == 5 or marker.type == 6 or marker.type == 8]

    async def get_tc_storage_contents(self, eid: int = None, combine_stacks: bool = False) -> RustContents:

        if eid is None:
            raise ValueError("EID cannot be None")

        returned_data = await self.get_entity_info(eid)

        target_time = datetime.utcfromtimestamp(int(returned_data.protectionExpiry))
        difference = target_time - datetime.utcnow()

        items = []

        for item in returned_data.items:
            items.append(RustItem(translate_id_to_stack(item.itemId), item.itemId, item.quantity, item.itemIsBlueprint))

        if combine_stacks:
            merged_map = defaultdict(tuple)

            for item in items:
                data = merged_map[str(item.itemId)]
                if data:
                    count = int(data[0]) + int(item.quantity)
                    merged_map[str(item.itemId)] = (count, bool(item.isBlueprint))
                else:
                    merged_map[str(item.itemId)] = (int(item.quantity), bool(item.isBlueprint))

            items = []
            for key in merged_map.keys():
                items.append(
                    RustItem(translate_id_to_stack(key), key, int(merged_map[key][0]), bool(merged_map[key][1])))

        return RustContents(difference, bool(returned_data.hasProtection), items)



from .exceptions import RateLimitError, ServerNotResponsiveError, CommandsNotEnabledError, ResponseNotRecievedError, \
    PrefixNotDefinedError, ImageError, ClientNotConnectedError, RequestError, SmartDeviceRegistrationError



class RustTeamInfo:

    def __init__(self, data) -> None:
        self.leaderSteamId: int = data.leaderSteamId
        self.members = [RustTeamMember(member) for member in data.members]
        self.mapNotes = [RustTeamNote(note) for note in data.mapNotes]
        self.leaderMapNotes = [RustTeamNote(note) for note in data.leaderMapNotes]

    def __str__(self) -> str:
        return "RustTeamInfo[leaderSteamId={}, members={}, mapNotes={}, leaderMapNotes={}]".format(self.leaderSteamId,
                                                                                                   self.members,
                                                                                                   self.mapNotes,
                                                                                                   self.leaderMapNotes)


class RustTeamMember:

    def __init__(self, data) -> None:
        self.steamId: int = data.steamId
        self.name: str = data.name
        self.x: float = data.x
        self.y: float = data.y
        self.isOnline: bool = data.isOnline
        self.spawnTime: int = data.spawnTime
        self.isAlive: bool = data.isAlive
        self.deathTime: int = data.deathTime

    def __str__(self) -> str:
        return "RustTeamMember[steamId={}, name={}, x={}, y={}, isOnline={}, spawnTime={}, isAlive={}, deathTime={}]".format(
            self.steamId, self.name, self.x, self.y, self.isOnline, self.spawnTime, self.isAlive, self.deathTime)


class RustTeamNote:

    def __init__(self, data) -> None:
        self.type: int = data.type
        self.x: float = data.x
        self.y: float = data.y

    def __str__(self) -> str:
        return "RustTeamNote[type={}, x={}, y={}]".format(self.type, self.x, self.y)



class RustInfo:

    def __init__(self, data) -> None:
        self.url: str = data.url
        self.name: str = data.name
        self.map: str = data.map
        self.size: int = data.mapSize
        self.players: int = data.players
        self.max_players: int = data.maxPlayers
        self.queued_players: int = data.queuedPlayers
        self.seed: int = data.seed

    def __str__(self) -> str:
        return "RustInfo[url={}, name={}, map={}, size={}, players={}, max_players={}, queued_players={}, seed={}]".format(
            self.url, self.name, self.map, self.size, self.players, self.max_players, self.queued_players, self.seed)



import asyncio
from asyncio.futures import Future
from datetime import datetime

from . import Command, CommandTime
from ..api.structures import RustChatMessage
from ..commands.command_options import CommandOptions
from ..utils import RegisteredListener


class CommandHandler:

    def __init__(self, command_options: CommandOptions) -> None:
        self.command_options = command_options

    def register_command(self, command, data) -> None:

        if not asyncio.iscoroutinefunction(data[0]):
            raise TypeError("The event registered must be a coroutine")

        setattr(self, command, data)

    def _schedule_event(self, loop, coro, arg) -> None:

        def callback(inner_future: Future):
            inner_future.result()

        future: Future = asyncio.run_coroutine_threadsafe(coro(arg), loop)
        future.add_done_callback(callback)

    def run_command(self, message: RustChatMessage, prefix) -> None:

        if prefix == self.command_options.prefix:
            command = message.message.split(" ")[0][len(prefix):]
        else:
            command = prefix

        if hasattr(self, command):
            coro, loop = getattr(self, command)

            time = CommandTime(datetime.utcfromtimestamp(message.time), message.time)

            self._schedule_event(loop, coro,
                                 Command(message.name, message.steamId, time, command, message.message.split(" ")[1:]))

    def has_command(self, listener: RegisteredListener) -> bool:
        return hasattr(self, listener.listener_id)

    def remove_command(self, listener: RegisteredListener) -> None:
        delattr(self, listener.listener_id)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rustplus.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0erustplus.proto\x12\x08rustplus\"`\n\nAppMessage\x12\'\n\x08response\x18\x01 \x01(\x0b\x32\x15.rustplus.AppResponse\x12)\n\tbroadcast\x18\x02 \x01(\x0b\x32\x16.rustplus.AppBroadcast\"\x9b\x05\n\nAppRequest\x12\x0b\n\x03seq\x18\x01 \x02(\r\x12\x10\n\x08playerId\x18\x02 \x02(\x04\x12\x13\n\x0bplayerToken\x18\x03 \x02(\x05\x12\x10\n\x08\x65ntityId\x18\x04 \x01(\r\x12#\n\x07getInfo\x18\x08 \x01(\x0b\x32\x12.rustplus.AppEmpty\x12#\n\x07getTime\x18\t \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\"\n\x06getMap\x18\n \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\'\n\x0bgetTeamInfo\x18\x0b \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\'\n\x0bgetTeamChat\x18\x0c \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\x31\n\x0fsendTeamMessage\x18\r \x01(\x0b\x32\x18.rustplus.AppSendMessage\x12)\n\rgetEntityInfo\x18\x0e \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\x33\n\x0esetEntityValue\x18\x0f \x01(\x0b\x32\x1b.rustplus.AppSetEntityValue\x12-\n\x11\x63heckSubscription\x18\x10 \x01(\x0b\x32\x12.rustplus.AppEmpty\x12*\n\x0fsetSubscription\x18\x11 \x01(\x0b\x32\x11.rustplus.AppFlag\x12)\n\rgetMapMarkers\x18\x12 \x01(\x0b\x32\x12.rustplus.AppEmpty\x12\x37\n\x0egetCameraFrame\x18\x13 \x01(\x0b\x32\x1f.rustplus.AppCameraFrameRequest\x12\x35\n\x0fpromoteToLeader\x18\x14 \x01(\x0b\x32\x1c.rustplus.AppPromoteToLeader\"!\n\x0e\x41ppSendMessage\x12\x0f\n\x07message\x18\x01 \x02(\t\"\"\n\x11\x41ppSetEntityValue\x12\r\n\x05value\x18\x01 \x02(\x08\":\n\x15\x41ppCameraFrameRequest\x12\x12\n\nidentifier\x18\x01 \x02(\t\x12\r\n\x05\x66rame\x18\x02 \x02(\r\"%\n\x12\x41ppPromoteToLeader\x12\x0f\n\x07steamId\x18\x01 \x02(\x04\"\xc1\x03\n\x0b\x41ppResponse\x12\x0b\n\x03seq\x18\x01 \x02(\r\x12%\n\x07success\x18\x04 \x01(\x0b\x32\x14.rustplus.AppSuccess\x12!\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x12.rustplus.AppError\x12\x1f\n\x04info\x18\x06 \x01(\x0b\x32\x11.rustplus.AppInfo\x12\x1f\n\x04time\x18\x07 \x01(\x0b\x32\x11.rustplus.AppTime\x12\x1d\n\x03map\x18\x08 \x01(\x0b\x32\x10.rustplus.AppMap\x12\'\n\x08teamInfo\x18\t \x01(\x0b\x32\x15.rustplus.AppTeamInfo\x12\'\n\x08teamChat\x18\n \x01(\x0b\x32\x15.rustplus.AppTeamChat\x12+\n\nentityInfo\x18\x0b \x01(\x0b\x32\x17.rustplus.AppEntityInfo\x12\x1f\n\x04\x66lag\x18\x0c \x01(\x0b\x32\x11.rustplus.AppFlag\x12+\n\nmapMarkers\x18\r \x01(\x0b\x32\x17.rustplus.AppMapMarkers\x12-\n\x0b\x63\x61meraFrame\x18\x0e \x01(\x0b\x32\x18.rustplus.AppCameraFrame\"\x9f\x01\n\x0c\x41ppBroadcast\x12-\n\x0bteamChanged\x18\x04 \x01(\x0b\x32\x18.rustplus.AppTeamChanged\x12-\n\x0bteamMessage\x18\x05 \x01(\x0b\x32\x18.rustplus.AppTeamMessage\x12\x31\n\rentityChanged\x18\x06 \x01(\x0b\x32\x1a.rustplus.AppEntityChanged\"\n\n\x08\x41ppEmpty\"\x0c\n\nAppSuccess\"\x19\n\x08\x41ppError\x12\r\n\x05\x65rror\x18\x01 \x02(\t\"\xc1\x01\n\x07\x41ppInfo\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x13\n\x0bheaderImage\x18\x02 \x02(\t\x12\x0b\n\x03url\x18\x03 \x02(\t\x12\x0b\n\x03map\x18\x04 \x02(\t\x12\x0f\n\x07mapSize\x18\x05 \x02(\r\x12\x10\n\x08wipeTime\x18\x06 \x02(\r\x12\x0f\n\x07players\x18\x07 \x02(\r\x12\x12\n\nmaxPlayers\x18\x08 \x02(\r\x12\x15\n\rqueuedPlayers\x18\t \x02(\r\x12\x0c\n\x04seed\x18\n \x01(\r\x12\x0c\n\x04salt\x18\x0b \x01(\r\"e\n\x07\x41ppTime\x12\x18\n\x10\x64\x61yLengthMinutes\x18\x01 \x02(\x02\x12\x11\n\ttimeScale\x18\x02 \x02(\x02\x12\x0f\n\x07sunrise\x18\x03 \x02(\x02\x12\x0e\n\x06sunset\x18\x04 \x02(\x02\x12\x0c\n\x04time\x18\x05 \x02(\x02\"\xc1\x01\n\x06\x41ppMap\x12\r\n\x05width\x18\x01 \x02(\r\x12\x0e\n\x06height\x18\x02 \x02(\r\x12\x10\n\x08jpgImage\x18\x03 \x02(\x0c\x12\x13\n\x0boceanMargin\x18\x04 \x02(\x05\x12,\n\tmonuments\x18\x05 \x03(\x0b\x32\x19.rustplus.AppMap.Monument\x12\x12\n\nbackground\x18\x06 \x01(\t\x1a/\n\x08Monument\x12\r\n\x05token\x18\x01 \x02(\t\x12\t\n\x01x\x18\x02 \x02(\x02\x12\t\n\x01y\x18\x03 \x02(\x02\"\xea\x02\n\x0b\x41ppTeamInfo\x12\x15\n\rleaderSteamId\x18\x01 \x02(\x04\x12-\n\x07members\x18\x02 \x03(\x0b\x32\x1c.rustplus.AppTeamInfo.Member\x12,\n\x08mapNotes\x18\x03 \x03(\x0b\x32\x1a.rustplus.AppTeamInfo.Note\x12\x32\n\x0eleaderMapNotes\x18\x04 \x03(\x0b\x32\x1a.rustplus.AppTeamInfo.Note\x1a\x86\x01\n\x06Member\x12\x0f\n\x07steamId\x18\x01 \x02(\x04\x12\x0c\n\x04name\x18\x02 \x02(\t\x12\t\n\x01x\x18\x03 \x02(\x02\x12\t\n\x01y\x18\x04 \x02(\x02\x12\x10\n\x08isOnline\x18\x05 \x02(\x08\x12\x11\n\tspawnTime\x18\x06 \x02(\r\x12\x0f\n\x07isAlive\x18\x07 \x02(\x08\x12\x11\n\tdeathTime\x18\x08 \x02(\r\x1a*\n\x04Note\x12\x0c\n\x04type\x18\x02 \x02(\x05\x12\t\n\x01x\x18\x03 \x02(\x02\x12\t\n\x01y\x18\x04 \x02(\x02\"9\n\x0b\x41ppTeamChat\x12*\n\x08messages\x18\x01 \x03(\x0b\x32\x18.rustplus.AppChatMessage\"]\n\x0e\x41ppChatMessage\x12\x0f\n\x07steamId\x18\x01 \x02(\x04\x12\x0c\n\x04name\x18\x02 \x02(\t\x12\x0f\n\x07message\x18\x03 \x02(\t\x12\r\n\x05\x63olor\x18\x04 \x02(\t\x12\x0c\n\x04time\x18\x05 \x02(\r\"c\n\rAppEntityInfo\x12%\n\x04type\x18\x01 \x02(\x0e\x32\x17.rustplus.AppEntityType\x12+\n\x07payload\x18\x03 \x02(\x0b\x32\x1a.rustplus.AppEntityPayload\"\xd7\x01\n\x10\x41ppEntityPayload\x12\r\n\x05value\x18\x01 \x01(\x08\x12.\n\x05items\x18\x02 \x03(\x0b\x32\x1f.rustplus.AppEntityPayload.Item\x12\x10\n\x08\x63\x61pacity\x18\x03 \x01(\x05\x12\x15\n\rhasProtection\x18\x04 \x01(\x08\x12\x18\n\x10protectionExpiry\x18\x05 \x01(\r\x1a\x41\n\x04Item\x12\x0e\n\x06itemId\x18\x01 \x02(\x05\x12\x10\n\x08quantity\x18\x02 \x02(\x05\x12\x17\n\x0fitemIsBlueprint\x18\x03 \x02(\x08\"\x18\n\x07\x41ppFlag\x12\r\n\x05value\x18\x01 \x02(\x08\"5\n\rAppMapMarkers\x12$\n\x07markers\x18\x01 \x03(\x0b\x32\x13.rustplus.AppMarker\"5\n\x07Vector4\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x12\t\n\x01w\x18\x04 \x01(\x02\"\xc3\x03\n\tAppMarker\x12\n\n\x02id\x18\x01 \x02(\r\x12%\n\x04type\x18\x02 \x02(\x0e\x32\x17.rustplus.AppMarkerType\x12\t\n\x01x\x18\x03 \x02(\x02\x12\t\n\x01y\x18\x04 \x02(\x02\x12\x0f\n\x07steamId\x18\x05 \x01(\x04\x12\x10\n\x08rotation\x18\x06 \x01(\x02\x12\x0e\n\x06radius\x18\x07 \x01(\x02\x12!\n\x06\x63olor1\x18\x08 \x01(\x0b\x32\x11.rustplus.Vector4\x12!\n\x06\x63olor2\x18\t \x01(\x0b\x32\x11.rustplus.Vector4\x12\r\n\x05\x61lpha\x18\n \x01(\x02\x12\x0c\n\x04name\x18\x0b \x01(\t\x12\x31\n\nsellOrders\x18\r \x03(\x0b\x32\x1d.rustplus.AppMarker.SellOrder\x1a\xa3\x01\n\tSellOrder\x12\x0e\n\x06itemId\x18\x01 \x02(\x05\x12\x10\n\x08quantity\x18\x02 \x02(\x05\x12\x12\n\ncurrencyId\x18\x03 \x02(\x05\x12\x13\n\x0b\x63ostPerItem\x18\x04 \x02(\x05\x12\x15\n\ramountInStock\x18\x05 \x02(\x05\x12\x17\n\x0fitemIsBlueprint\x18\x06 \x02(\x08\x12\x1b\n\x13\x63urrencyIsBlueprint\x18\x07 \x02(\x08\"1\n\x0e\x41ppCameraFrame\x12\r\n\x05\x66rame\x18\x01 \x02(\r\x12\x10\n\x08jpgImage\x18\x02 \x02(\x0c\"K\n\x0e\x41ppTeamChanged\x12\x10\n\x08playerId\x18\x01 \x02(\x04\x12\'\n\x08teamInfo\x18\x02 \x02(\x0b\x32\x15.rustplus.AppTeamInfo\";\n\x0e\x41ppTeamMessage\x12)\n\x07message\x18\x01 \x02(\x0b\x32\x18.rustplus.AppChatMessage\"Q\n\x10\x41ppEntityChanged\x12\x10\n\x08\x65ntityId\x18\x01 \x02(\r\x12+\n\x07payload\x18\x02 \x02(\x0b\x32\x1a.rustplus.AppEntityPayload*:\n\rAppEntityType\x12\n\n\x06Switch\x10\x01\x12\t\n\x05\x41larm\x10\x02\x12\x12\n\x0eStorageMonitor\x10\x03*\x8b\x01\n\rAppMarkerType\x12\n\n\x06Player\x10\x01\x12\r\n\tExplosion\x10\x02\x12\x12\n\x0eVendingMachine\x10\x03\x12\x08\n\x04\x43H47\x10\x04\x12\r\n\tCargoShip\x10\x05\x12\t\n\x05\x43rate\x10\x06\x12\x11\n\rGenericRadius\x10\x07\x12\x14\n\x10PatrolHelicopter\x10\x08')

_APPENTITYTYPE = DESCRIPTOR.enum_types_by_name['AppEntityType']
AppEntityType = enum_type_wrapper.EnumTypeWrapper(_APPENTITYTYPE)
_APPMARKERTYPE = DESCRIPTOR.enum_types_by_name['AppMarkerType']
AppMarkerType = enum_type_wrapper.EnumTypeWrapper(_APPMARKERTYPE)
Switch = 1
Alarm = 2
StorageMonitor = 3
Player = 1
Explosion = 2
VendingMachine = 3
CH47 = 4
CargoShip = 5
Crate = 6
GenericRadius = 7
PatrolHelicopter = 8


_APPMESSAGE = DESCRIPTOR.message_types_by_name['AppMessage']
_APPREQUEST = DESCRIPTOR.message_types_by_name['AppRequest']
_APPSENDMESSAGE = DESCRIPTOR.message_types_by_name['AppSendMessage']
_APPSETENTITYVALUE = DESCRIPTOR.message_types_by_name['AppSetEntityValue']
_APPCAMERAFRAMEREQUEST = DESCRIPTOR.message_types_by_name['AppCameraFrameRequest']
_APPPROMOTETOLEADER = DESCRIPTOR.message_types_by_name['AppPromoteToLeader']
_APPRESPONSE = DESCRIPTOR.message_types_by_name['AppResponse']
_APPBROADCAST = DESCRIPTOR.message_types_by_name['AppBroadcast']
_APPEMPTY = DESCRIPTOR.message_types_by_name['AppEmpty']
_APPSUCCESS = DESCRIPTOR.message_types_by_name['AppSuccess']
_APPERROR = DESCRIPTOR.message_types_by_name['AppError']
_APPINFO = DESCRIPTOR.message_types_by_name['AppInfo']
_APPTIME = DESCRIPTOR.message_types_by_name['AppTime']
_APPMAP = DESCRIPTOR.message_types_by_name['AppMap']
_APPMAP_MONUMENT = _APPMAP.nested_types_by_name['Monument']
_APPTEAMINFO = DESCRIPTOR.message_types_by_name['AppTeamInfo']
_APPTEAMINFO_MEMBER = _APPTEAMINFO.nested_types_by_name['Member']
_APPTEAMINFO_NOTE = _APPTEAMINFO.nested_types_by_name['Note']
_APPTEAMCHAT = DESCRIPTOR.message_types_by_name['AppTeamChat']
_APPCHATMESSAGE = DESCRIPTOR.message_types_by_name['AppChatMessage']
_APPENTITYINFO = DESCRIPTOR.message_types_by_name['AppEntityInfo']
_APPENTITYPAYLOAD = DESCRIPTOR.message_types_by_name['AppEntityPayload']
_APPENTITYPAYLOAD_ITEM = _APPENTITYPAYLOAD.nested_types_by_name['Item']
_APPFLAG = DESCRIPTOR.message_types_by_name['AppFlag']
_APPMAPMARKERS = DESCRIPTOR.message_types_by_name['AppMapMarkers']
_VECTOR4 = DESCRIPTOR.message_types_by_name['Vector4']
_APPMARKER = DESCRIPTOR.message_types_by_name['AppMarker']
_APPMARKER_SELLORDER = _APPMARKER.nested_types_by_name['SellOrder']
_APPCAMERAFRAME = DESCRIPTOR.message_types_by_name['AppCameraFrame']
_APPTEAMCHANGED = DESCRIPTOR.message_types_by_name['AppTeamChanged']
_APPTEAMMESSAGE = DESCRIPTOR.message_types_by_name['AppTeamMessage']
_APPENTITYCHANGED = DESCRIPTOR.message_types_by_name['AppEntityChanged']
AppMessage = _reflection.GeneratedProtocolMessageType('AppMessage', (_message.Message,), {
  'DESCRIPTOR' : _APPMESSAGE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppMessage)
  })
_sym_db.RegisterMessage(AppMessage)

AppRequest = _reflection.GeneratedProtocolMessageType('AppRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPREQUEST,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppRequest)
  })
_sym_db.RegisterMessage(AppRequest)

AppSendMessage = _reflection.GeneratedProtocolMessageType('AppSendMessage', (_message.Message,), {
  'DESCRIPTOR' : _APPSENDMESSAGE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppSendMessage)
  })
_sym_db.RegisterMessage(AppSendMessage)

AppSetEntityValue = _reflection.GeneratedProtocolMessageType('AppSetEntityValue', (_message.Message,), {
  'DESCRIPTOR' : _APPSETENTITYVALUE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppSetEntityValue)
  })
_sym_db.RegisterMessage(AppSetEntityValue)

AppCameraFrameRequest = _reflection.GeneratedProtocolMessageType('AppCameraFrameRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPCAMERAFRAMEREQUEST,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppCameraFrameRequest)
  })
_sym_db.RegisterMessage(AppCameraFrameRequest)

AppPromoteToLeader = _reflection.GeneratedProtocolMessageType('AppPromoteToLeader', (_message.Message,), {
  'DESCRIPTOR' : _APPPROMOTETOLEADER,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppPromoteToLeader)
  })
_sym_db.RegisterMessage(AppPromoteToLeader)

AppResponse = _reflection.GeneratedProtocolMessageType('AppResponse', (_message.Message,), {
  'DESCRIPTOR' : _APPRESPONSE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppResponse)
  })
_sym_db.RegisterMessage(AppResponse)

AppBroadcast = _reflection.GeneratedProtocolMessageType('AppBroadcast', (_message.Message,), {
  'DESCRIPTOR' : _APPBROADCAST,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppBroadcast)
  })
_sym_db.RegisterMessage(AppBroadcast)

AppEmpty = _reflection.GeneratedProtocolMessageType('AppEmpty', (_message.Message,), {
  'DESCRIPTOR' : _APPEMPTY,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppEmpty)
  })
_sym_db.RegisterMessage(AppEmpty)

AppSuccess = _reflection.GeneratedProtocolMessageType('AppSuccess', (_message.Message,), {
  'DESCRIPTOR' : _APPSUCCESS,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppSuccess)
  })
_sym_db.RegisterMessage(AppSuccess)

AppError = _reflection.GeneratedProtocolMessageType('AppError', (_message.Message,), {
  'DESCRIPTOR' : _APPERROR,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppError)
  })
_sym_db.RegisterMessage(AppError)

AppInfo = _reflection.GeneratedProtocolMessageType('AppInfo', (_message.Message,), {
  'DESCRIPTOR' : _APPINFO,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppInfo)
  })
_sym_db.RegisterMessage(AppInfo)

AppTime = _reflection.GeneratedProtocolMessageType('AppTime', (_message.Message,), {
  'DESCRIPTOR' : _APPTIME,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppTime)
  })
_sym_db.RegisterMessage(AppTime)

AppMap = _reflection.GeneratedProtocolMessageType('AppMap', (_message.Message,), {

  'Monument' : _reflection.GeneratedProtocolMessageType('Monument', (_message.Message,), {
    'DESCRIPTOR' : _APPMAP_MONUMENT,
    '__module__' : 'rustplus_pb2'
    # @@protoc_insertion_point(class_scope:rustplus.AppMap.Monument)
    })
  ,
  'DESCRIPTOR' : _APPMAP,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppMap)
  })
_sym_db.RegisterMessage(AppMap)
_sym_db.RegisterMessage(AppMap.Monument)

AppTeamInfo = _reflection.GeneratedProtocolMessageType('AppTeamInfo', (_message.Message,), {

  'Member' : _reflection.GeneratedProtocolMessageType('Member', (_message.Message,), {
    'DESCRIPTOR' : _APPTEAMINFO_MEMBER,
    '__module__' : 'rustplus_pb2'
    # @@protoc_insertion_point(class_scope:rustplus.AppTeamInfo.Member)
    })
  ,

  'Note' : _reflection.GeneratedProtocolMessageType('Note', (_message.Message,), {
    'DESCRIPTOR' : _APPTEAMINFO_NOTE,
    '__module__' : 'rustplus_pb2'
    # @@protoc_insertion_point(class_scope:rustplus.AppTeamInfo.Note)
    })
  ,
  'DESCRIPTOR' : _APPTEAMINFO,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppTeamInfo)
  })
_sym_db.RegisterMessage(AppTeamInfo)
_sym_db.RegisterMessage(AppTeamInfo.Member)
_sym_db.RegisterMessage(AppTeamInfo.Note)

AppTeamChat = _reflection.GeneratedProtocolMessageType('AppTeamChat', (_message.Message,), {
  'DESCRIPTOR' : _APPTEAMCHAT,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppTeamChat)
  })
_sym_db.RegisterMessage(AppTeamChat)

AppChatMessage = _reflection.GeneratedProtocolMessageType('AppChatMessage', (_message.Message,), {
  'DESCRIPTOR' : _APPCHATMESSAGE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppChatMessage)
  })
_sym_db.RegisterMessage(AppChatMessage)

AppEntityInfo = _reflection.GeneratedProtocolMessageType('AppEntityInfo', (_message.Message,), {
  'DESCRIPTOR' : _APPENTITYINFO,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppEntityInfo)
  })
_sym_db.RegisterMessage(AppEntityInfo)

AppEntityPayload = _reflection.GeneratedProtocolMessageType('AppEntityPayload', (_message.Message,), {

  'Item' : _reflection.GeneratedProtocolMessageType('Item', (_message.Message,), {
    'DESCRIPTOR' : _APPENTITYPAYLOAD_ITEM,
    '__module__' : 'rustplus_pb2'
    # @@protoc_insertion_point(class_scope:rustplus.AppEntityPayload.Item)
    })
  ,
  'DESCRIPTOR' : _APPENTITYPAYLOAD,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppEntityPayload)
  })
_sym_db.RegisterMessage(AppEntityPayload)
_sym_db.RegisterMessage(AppEntityPayload.Item)

AppFlag = _reflection.GeneratedProtocolMessageType('AppFlag', (_message.Message,), {
  'DESCRIPTOR' : _APPFLAG,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppFlag)
  })
_sym_db.RegisterMessage(AppFlag)

AppMapMarkers = _reflection.GeneratedProtocolMessageType('AppMapMarkers', (_message.Message,), {
  'DESCRIPTOR' : _APPMAPMARKERS,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppMapMarkers)
  })
_sym_db.RegisterMessage(AppMapMarkers)

Vector4 = _reflection.GeneratedProtocolMessageType('Vector4', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR4,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.Vector4)
  })
_sym_db.RegisterMessage(Vector4)

AppMarker = _reflection.GeneratedProtocolMessageType('AppMarker', (_message.Message,), {

  'SellOrder' : _reflection.GeneratedProtocolMessageType('SellOrder', (_message.Message,), {
    'DESCRIPTOR' : _APPMARKER_SELLORDER,
    '__module__' : 'rustplus_pb2'
    # @@protoc_insertion_point(class_scope:rustplus.AppMarker.SellOrder)
    })
  ,
  'DESCRIPTOR' : _APPMARKER,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppMarker)
  })
_sym_db.RegisterMessage(AppMarker)
_sym_db.RegisterMessage(AppMarker.SellOrder)

AppCameraFrame = _reflection.GeneratedProtocolMessageType('AppCameraFrame', (_message.Message,), {
  'DESCRIPTOR' : _APPCAMERAFRAME,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppCameraFrame)
  })
_sym_db.RegisterMessage(AppCameraFrame)

AppTeamChanged = _reflection.GeneratedProtocolMessageType('AppTeamChanged', (_message.Message,), {
  'DESCRIPTOR' : _APPTEAMCHANGED,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppTeamChanged)
  })
_sym_db.RegisterMessage(AppTeamChanged)

AppTeamMessage = _reflection.GeneratedProtocolMessageType('AppTeamMessage', (_message.Message,), {
  'DESCRIPTOR' : _APPTEAMMESSAGE,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppTeamMessage)
  })
_sym_db.RegisterMessage(AppTeamMessage)

AppEntityChanged = _reflection.GeneratedProtocolMessageType('AppEntityChanged', (_message.Message,), {
  'DESCRIPTOR' : _APPENTITYCHANGED,
  '__module__' : 'rustplus_pb2'
  # @@protoc_insertion_point(class_scope:rustplus.AppEntityChanged)
  })
_sym_db.RegisterMessage(AppEntityChanged)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _APPENTITYTYPE._serialized_start=3828
  _APPENTITYTYPE._serialized_end=3886
  _APPMARKERTYPE._serialized_start=3889
  _APPMARKERTYPE._serialized_end=4028
  _APPMESSAGE._serialized_start=28
  _APPMESSAGE._serialized_end=124
  _APPREQUEST._serialized_start=127
  _APPREQUEST._serialized_end=794
  _APPSENDMESSAGE._serialized_start=796
  _APPSENDMESSAGE._serialized_end=829
  _APPSETENTITYVALUE._serialized_start=831
  _APPSETENTITYVALUE._serialized_end=865
  _APPCAMERAFRAMEREQUEST._serialized_start=867
  _APPCAMERAFRAMEREQUEST._serialized_end=925
  _APPPROMOTETOLEADER._serialized_start=927
  _APPPROMOTETOLEADER._serialized_end=964
  _APPRESPONSE._serialized_start=967
  _APPRESPONSE._serialized_end=1416
  _APPBROADCAST._serialized_start=1419
  _APPBROADCAST._serialized_end=1578
  _APPEMPTY._serialized_start=1580
  _APPEMPTY._serialized_end=1590
  _APPSUCCESS._serialized_start=1592
  _APPSUCCESS._serialized_end=1604
  _APPERROR._serialized_start=1606
  _APPERROR._serialized_end=1631
  _APPINFO._serialized_start=1634
  _APPINFO._serialized_end=1827
  _APPTIME._serialized_start=1829
  _APPTIME._serialized_end=1930
  _APPMAP._serialized_start=1933
  _APPMAP._serialized_end=2126
  _APPMAP_MONUMENT._serialized_start=2079
  _APPMAP_MONUMENT._serialized_end=2126
  _APPTEAMINFO._serialized_start=2129
  _APPTEAMINFO._serialized_end=2491
  _APPTEAMINFO_MEMBER._serialized_start=2313
  _APPTEAMINFO_MEMBER._serialized_end=2447
  _APPTEAMINFO_NOTE._serialized_start=2449
  _APPTEAMINFO_NOTE._serialized_end=2491
  _APPTEAMCHAT._serialized_start=2493
  _APPTEAMCHAT._serialized_end=2550
  _APPCHATMESSAGE._serialized_start=2552
  _APPCHATMESSAGE._serialized_end=2645
  _APPENTITYINFO._serialized_start=2647
  _APPENTITYINFO._serialized_end=2746
  _APPENTITYPAYLOAD._serialized_start=2749
  _APPENTITYPAYLOAD._serialized_end=2964
  _APPENTITYPAYLOAD_ITEM._serialized_start=2899
  _APPENTITYPAYLOAD_ITEM._serialized_end=2964
  _APPFLAG._serialized_start=2966
  _APPFLAG._serialized_end=2990
  _APPMAPMARKERS._serialized_start=2992
  _APPMAPMARKERS._serialized_end=3045
  _VECTOR4._serialized_start=3047
  _VECTOR4._serialized_end=3100
  _APPMARKER._serialized_start=3103
  _APPMARKER._serialized_end=3554
  _APPMARKER_SELLORDER._serialized_start=3391
  _APPMARKER_SELLORDER._serialized_end=3554
  _APPCAMERAFRAME._serialized_start=3556
  _APPCAMERAFRAME._serialized_end=3605
  _APPTEAMCHANGED._serialized_start=3607
  _APPTEAMCHANGED._serialized_end=3682
  _APPTEAMMESSAGE._serialized_start=3684
  _APPTEAMMESSAGE._serialized_end=3743
  _APPENTITYCHANGED._serialized_start=3745
  _APPENTITYCHANGED._serialized_end=3826
# @@protoc_insertion_point(module_scope)



class RustTime:

    def __init__(self, day_length, sunrise, sunset, time, raw_time) -> None:
        self.day_length: float = day_length
        self.sunrise: str = sunrise
        self.sunset: str = sunset
        self.time: str = time
        self.raw_time: float = raw_time

    def __str__(self) -> str:
        return f"RustTime[day_length={self.day_length}, sunrise={self.sunrise}, sunset={self.sunset}, time={self.time}, raw_time={self.raw_time}]"



import logging
import time
from datetime import datetime
from threading import Thread
from typing import Optional
import websocket

from .rustplus_pb2 import AppMessage, AppRequest
from ..structures import RustChatMessage
from ...exceptions import ClientNotConnectedError


class RustWebsocket(websocket.WebSocket):

    def __init__(self, ip, port, remote, use_proxy):

        self.ip = ip
        self.port = port
        self.thread = None
        self.open = False
        self.use_proxy = use_proxy
        self.remote = remote
        self.logger = logging.getLogger("rustplus.py")
        self.connected_time = time.time()

        super().__init__(enable_multithread=True)

    def connect(self, ignore=False) -> None:

        if ((not self.open) or ignore) and not self.remote.is_pending:

            while True:

                self.remote.is_pending = True

                try:
                    address = f"wss://companion-rust.facepunch.com/game/{self.ip}/{self.port}" if self.use_proxy else \
                        f"ws://{self.ip}:{self.port} "
                    super().connect(address)
                    self.connected_time = time.time()
                    break
                except Exception:
                    self.logger.warning(
                        f"{datetime.now().strftime('%d/%m/%Y %H:%M:%S')} "
                        "[RustPlus.py] Cannot Connect to server. Retrying in 20 seconds")
                    time.sleep(20)

            self.remote.is_pending = False

            self.open = True

            self.thread = Thread(target=self.run, name="[RustPlus.py] WebsocketThread", daemon=True)
            self.thread.start()

    def close(self) -> None:

        super().close()
        self.open = False

    def send_message(self, message: AppRequest) -> None:
        """
        Send the Protobuf to the server
        """
        try:
            self.remote.pending_requests[message.seq] = message
            self.send_binary(message.SerializeToString())
        except Exception:
            if not self.open:
                raise ClientNotConnectedError("Not Connected")

    def run(self) -> None:

        while self.open:
            try:
                data = self.recv()
                app_message = AppMessage()
                app_message.ParseFromString(data)

                try:
                    del self.remote.pending_requests[app_message.response.seq]
                except KeyError:
                    pass

                self.handle_message(app_message)

            except Exception:
                if self.open:
                    self.logger.warning(
                        f"{datetime.now().strftime('%d/%m/%Y %H:%M:%S')} [RustPlus.py] Connection interrupted, Retrying")
                    self.connect(ignore=True)
                    return
                return

    def handle_message(self, app_message: AppMessage) -> None:

        if app_message.response.seq in self.remote.ignored_responses:
            self.remote.ignored_responses.remove(app_message.response.seq)
            return

        prefix = self.get_prefix(str(app_message.broadcast.teamMessage.message.message))

        if prefix is not None:

            message = RustChatMessage(app_message.broadcast.teamMessage.message)

            self.remote.command_handler.run_command(message, prefix)

        elif self.is_entity_broadcast(app_message):
            self.remote.event_handler.run_entity_event(app_message.broadcast.entityChanged.entityId, app_message)

        elif self.is_team_broadcast(app_message):
            self.remote.event_handler.run_team_event(app_message)

        elif self.is_message(app_message):
            self.remote.event_handler.run_chat_event(app_message)

        else:
            self.remote.responses[app_message.response.seq] = app_message

    def get_prefix(self, message: str) -> Optional[str]:
        if self.remote.use_commands:
            if message.startswith(self.remote.command_options.prefix):
                return self.remote.command_options.prefix
        else:
            return None

        for overrule in self.remote.command_options.overruling_commands:
            if message.startswith(overrule):
                return overrule

        return None

    @staticmethod
    def is_message(app_message) -> bool:
        return str(app_message.broadcast.teamMessage.message.message) != ""

    @staticmethod
    def is_entity_broadcast(app_message) -> bool:
        return str(app_message.broadcast.entityChanged) != ""

    @staticmethod
    def is_team_broadcast(app_message) -> bool:
        return str(app_message.broadcast.teamChanged) != ""

    async def _retry_failed_request(self, app_request: AppRequest):
        """
        Resends an AppRequest to the server if it has failed
        """
        await self.send_message(app_request)

    @staticmethod
    def _get_proto_cost(app_request) -> int:
        """
        Gets the cost of an AppRequest
        """
        costs = {"getTime": 1, "sendTeamMessage": 2, "getInfo": 1, "getTeamChat": 1, "getTeamInfo": 1,
                 "getMapMarkers": 1, "getMap": 5, "setEntityValue": 1, "getEntityInfo": 1, "promoteToLeader": 1}
        for request, cost in costs.items():
            if app_request.HasField(request):
                return cost

        raise ValueError()

    @staticmethod
    def error_present(message) -> bool:
        """
        Checks message for error
        """
        return message != ""



from datetime import timedelta
from typing import List

from .rust_item import RustItem


class RustContents:

    def __init__(self, protection_time, has_protection, contents) -> None:
        self.protectionTime: timedelta = protection_time
        self.hasProtection: bool = has_protection
        self.contents: List[RustItem] = contents

    def __str__(self) -> str:
        return "RustContents[protectionTime={}, hasProtection={}, contents={}]".format(self.protectionTime,
                                                                                       self.hasProtection,
                                                                                       self.contents)



class RustChatMessage:

    def __init__(self, data):
        self.steamId: int = data.steamId
        self.name: str = data.name
        self.message: str = data.message
        self.colour: str = data.color
        self.time: int = data.time

    def __repr__(self):
        return "RustChatMessage[steamId={}, senderName={}, message={}, colour={}, time={}]".format(self.steamId,
                                                                                                   self.name,
                                                                                                   self.message,
                                                                                                   self.colour,
                                                                                                   self.time)
